import yaml
import logging
import time
from typing import List
from google.protobuf.json_format import MessageToDict
from meshtastic import mt_config
from meshtastic.util import camel_to_snake, snake_to_camel, fromStr

# defs are from meshtastic/python/main


def _is_repeated_field(field_desc) -> bool:
    """Return True if the protobuf field is repeated.

    Protobuf 6.31.0+ exposes `is_repeated`, while older versions require
    checking `label == LABEL_REPEATED`.
    """
    if hasattr(field_desc, "is_repeated"):
        return bool(field_desc.is_repeated)
    return field_desc.label == field_desc.LABEL_REPEATED


def traverseConfig(config_root, config, interface_config) -> bool:
    """Iterate through current config level preferences and either traverse deeper if preference is a dict or set preference"""
    snake_name = camel_to_snake(config_root)
    for pref in config:
        pref_name = f"{snake_name}.{pref}"
        if isinstance(config[pref], dict):
            traverseConfig(pref_name, config[pref], interface_config)
        else:
            setPref(interface_config, pref_name, config[pref])

    return True


def splitCompoundName(comp_name: str) -> List[str]:
    """Split compound (dot separated) preference name into parts"""
    name: List[str] = comp_name.split(".")
    if len(name) < 2:
        name[0] = comp_name
        name.append(comp_name)
    return name


def setPref(config, comp_name, raw_val) -> bool:
    """Set a channel or preferences value"""

    name = splitCompoundName(comp_name)

    snake_name = camel_to_snake(name[-1])
    camel_name = snake_to_camel(name[-1])
    uni_name = camel_name if mt_config.camel_case else snake_name
    logging.debug(f"snake_name:{snake_name}")
    logging.debug(f"camel_name:{camel_name}")

    objDesc = config.DESCRIPTOR
    config_part = config
    config_type = objDesc.fields_by_name.get(name[0])
    if config_type and config_type.message_type is not None:
        for name_part in name[1:-1]:
            part_snake_name = camel_to_snake((name_part))
            config_part = getattr(config, config_type.name)
            config_type = config_type.message_type.fields_by_name.get(part_snake_name)
    pref = None
    if config_type and config_type.message_type is not None:
        pref = config_type.message_type.fields_by_name.get(snake_name)
    # Others like ChannelSettings are standalone
    elif config_type:
        pref = config_type

    if (not pref) or (not config_type):
        return False

    if isinstance(raw_val, str):
        val = fromStr(raw_val)
    else:
        val = raw_val
    logging.debug(f"valStr:{raw_val} val:{val}")

    if snake_name == "wifi_psk" and len(str(raw_val)) < 8:
        logging.info(f"Warning: network.wifi_psk must be 8 or more characters.")
        return False

    enumType = pref.enum_type
    # pylint: disable=C0123
    if enumType and type(val) == str:
        # We've failed so far to convert this string into an enum, try to find it by reflection
        e = enumType.values_by_name.get(val)
        if e:
            val = e.number
        else:
            logging.info(f"{name[0]}.{uni_name} does not have an enum called {val}, so you can not set it.")
            logging.info(f"Choices in sorted order are:")
            names = []
            for f in enumType.values:
                # Note: We must use the value of the enum (regardless if camel or snake case)
                names.append(f"{f.name}")
            for temp_name in sorted(names):
                logging.info(f"    {temp_name}")
            return False

    # repeating fields need to be handled with append, not setattr
    if not _is_repeated_field(pref):
        try:
            if config_type.message_type is not None:
                config_values = getattr(config_part, config_type.name)
                setattr(config_values, pref.name, val)
            else:
                setattr(config_part, snake_name, val)
        except TypeError:
            # The setter didn't like our arg type guess try again as a string
            config_values = getattr(config_part, config_type.name)
            setattr(config_values, pref.name, str(val))
    elif type(val) == list:
        new_vals = [fromStr(x) for x in val]
        config_values = getattr(config, config_type.name)
        getattr(config_values, pref.name)[:] = new_vals
    else:
        config_values = getattr(config, config_type.name)
        if val == 0:
            # clear values
            logging.info(f"Clearing {pref.name} list")
            del getattr(config_values, pref.name)[:]
        else:
            logging.info(f"Adding '{raw_val}' to the {pref.name} list")
            cur_vals = [x for x in getattr(config_values, pref.name) if x not in [0, "", b""]]
            cur_vals.append(val)
            getattr(config_values, pref.name)[:] = cur_vals
        return True

    prefix = f"{'.'.join(name[0:-1])}." if config_type.message_type is not None else ""
    logging.info(f"Set {prefix}{uni_name} to {raw_val}")

    return True


def config_import(interface, filename):
    with open(filename, encoding="utf8") as file:
        configuration = yaml.safe_load(file)
        closeNow = True

        interface.getNode("^local", False).beginSettingsTransaction()

        if "owner" in configuration:
            logging.info(f"Setting device owner to {configuration['owner']}")
            waitForAckNak = True
            interface.getNode("^local", False).setOwner(configuration["owner"])
            time.sleep(0.5)

        if "owner_short" in configuration:
            logging.info(f"Setting device owner short to {configuration['owner_short']}")
            waitForAckNak = True
            interface.getNode("^local", False).setOwner(long_name=None, short_name=configuration["owner_short"])
            time.sleep(0.5)

        if "ownerShort" in configuration:
            logging.info(f"Setting device owner short to {configuration['ownerShort']}")
            waitForAckNak = True
            interface.getNode("^local", False).setOwner(long_name=None, short_name=configuration["ownerShort"])
            time.sleep(0.5)

        if "channel_url" in configuration:
            logging.info(f"Setting channel url to {configuration['channel_url']}")
            interface.getNode("^local").setURL(configuration["channel_url"])
            time.sleep(0.5)

        if "channelUrl" in configuration:
            logging.info(f"Setting channel url to {configuration['channelUrl']}")
            interface.getNode("^local").setURL(configuration["channelUrl"])
            time.sleep(0.5)

        if "location" in configuration:
            alt = 0
            lat = 0.0
            lon = 0.0
            localConfig = interface.localNode.localConfig

            if "alt" in configuration["location"]:
                alt = int(configuration["location"]["alt"] or 0)
                logging.info(f"Fixing altitude at {alt} meters")
            if "lat" in configuration["location"]:
                lat = float(configuration["location"]["lat"] or 0)
                logging.info(f"Fixing latitude at {lat} degrees")
            if "lon" in configuration["location"]:
                lon = float(configuration["location"]["lon"] or 0)
                logging.info(f"Fixing longitude at {lon} degrees")
            logging.info("Setting device position")
            interface.localNode.setFixedPosition(lat, lon, alt)
            time.sleep(0.5)

        if "config" in configuration:
            localConfig = interface.getNode("^local").localConfig
            for section in configuration["config"]:
                traverseConfig(section, configuration["config"][section], localConfig)
                interface.getNode("^local").writeConfig(camel_to_snake(section))
                time.sleep(0.5)

        if "module_config" in configuration:
            moduleConfig = interface.getNode("^local").moduleConfig
            for section in configuration["module_config"]:
                traverseConfig(
                    section,
                    configuration["module_config"][section],
                    moduleConfig,
                )
                interface.getNode("^local").writeConfig(camel_to_snake(section))
                time.sleep(0.5)

        interface.getNode("^local", False).commitSettingsTransaction()
        logging.info("Writing modified configuration to device")


def config_export(interface) -> str:
    """used in --export-config"""
    configObj = {}

    owner = interface.getLongName()
    owner_short = interface.getShortName()
    channel_url = interface.localNode.getURL()
    myinfo = interface.getMyNodeInfo()
    pos = myinfo.get("position")
    lat = None
    lon = None
    alt = None
    if pos:
        lat = pos.get("latitude")
        lon = pos.get("longitude")
        alt = pos.get("altitude")

    if owner:
        configObj["owner"] = owner
    if owner_short:
        configObj["owner_short"] = owner_short
    if channel_url:
        if mt_config.camel_case:
            configObj["channelUrl"] = channel_url
        else:
            configObj["channel_url"] = channel_url
    # lat and lon don't make much sense without the other (so fill with 0s), and alt isn't meaningful without both
    if lat or lon:
        configObj["location"] = {"lat": lat or float(0), "lon": lon or float(0)}
        if alt:
            configObj["location"]["alt"] = alt

    config = MessageToDict(interface.localNode.localConfig)  # checkme - Used as a dictionary here and a string below
    if config:
        # Convert inner keys to correct snake/camelCase
        prefs = {}
        for pref in config:
            if mt_config.camel_case:
                prefs[snake_to_camel(pref)] = config[pref]
            else:
                prefs[pref] = config[pref]
            # mark base64 encoded fields as such
            if pref == "security":
                if "privateKey" in prefs[pref]:
                    prefs[pref]["privateKey"] = "base64:" + prefs[pref]["privateKey"]
                if "publicKey" in prefs[pref]:
                    prefs[pref]["publicKey"] = "base64:" + prefs[pref]["publicKey"]
                if "adminKey" in prefs[pref]:
                    for i in range(len(prefs[pref]["adminKey"])):
                        prefs[pref]["adminKey"][i] = "base64:" + prefs[pref]["adminKey"][i]
        if mt_config.camel_case:
            configObj["config"] = config  # Identical command here and 2 lines below?
        else:
            configObj["config"] = config

    module_config = MessageToDict(interface.localNode.moduleConfig)
    if module_config:
        # Convert inner keys to correct snake/camelCase
        prefs = {}
        for pref in module_config:
            if len(module_config[pref]) > 0:
                prefs[pref] = module_config[pref]
        if mt_config.camel_case:
            configObj["module_config"] = prefs
        else:
            configObj["module_config"] = prefs

    config_txt = "# start of Meshtastic configure yaml\n"  # checkme - "config" (now changed to config_out)
    # was used as a string here and a Dictionary above
    config_txt += yaml.dump(configObj)

    # logging.info(config_txt)
    return config_txt



import logging
import os
import platform
import shutil
import time
import subprocess
import threading
from typing import Any, Dict, Optional
 # Debounce notification sounds so a burst of queued messages only plays once.
_SOUND_DEBOUNCE_SECONDS = 0.8
_sound_timer: Optional[threading.Timer] = None
_sound_timer_lock = threading.Lock()
_last_sound_request = 0.0


def schedule_notification_sound(delay: float = _SOUND_DEBOUNCE_SECONDS) -> None:
    """Schedule a notification sound after a short quiet period.

    If more messages arrive before the delay elapses, the timer is reset.
    This prevents playing a sound for each message when a backlog flushes.
    """
    global _sound_timer, _last_sound_request

    now = time.monotonic()
    with _sound_timer_lock:
        _last_sound_request = now

        # Cancel any previously scheduled sound.
        if _sound_timer is not None:
            try:
                _sound_timer.cancel()
            except Exception:
                pass
            _sound_timer = None

        def _fire(expected_request_time: float) -> None:
            # Only play if nothing newer has been scheduled.
            with _sound_timer_lock:
                if expected_request_time != _last_sound_request:
                    return
            play_sound()

        _sound_timer = threading.Timer(delay, _fire, args=(now,))
        _sound_timer.daemon = True
        _sound_timer.start()
from contact.utilities.utils import (
    refresh_node_list,
    add_new_message,
)
from contact.ui.contact_ui import (
    draw_packetlog_win,
    draw_node_list,
    draw_messages_window,
    draw_channel_list,
    add_notification,
)
from contact.utilities.db_handler import (
    save_message_to_db,
    maybe_store_nodeinfo_in_db,
    get_name_from_database,
    update_node_info_in_db,
)
import contact.ui.default_config as config

from contact.utilities.singleton import ui_state, interface_state, app_state, menu_state


def play_sound():
    try:
        system = platform.system()
        sound_path = None
        executable = None

        if system == "Darwin":  # macOS
            sound_path = "/System/Library/Sounds/Ping.aiff"
            executable = "afplay"

        elif system == "Linux":
            ogg_path = "/usr/share/sounds/freedesktop/stereo/complete.oga"
            wav_path = "/usr/share/sounds/alsa/Front_Center.wav"  # common fallback

            if shutil.which("paplay") and os.path.exists(ogg_path):
                executable = "paplay"
                sound_path = ogg_path
            elif shutil.which("ffplay") and os.path.exists(ogg_path):
                executable = "ffplay"
                sound_path = ogg_path
            elif shutil.which("aplay") and os.path.exists(wav_path):
                executable = "aplay"
                sound_path = wav_path
            else:
                logging.warning("No suitable sound player or sound file found on Linux")

        if executable and sound_path:
            cmd = [executable, sound_path]
            if executable == "ffplay":
                cmd = [executable, "-nodisp", "-autoexit", sound_path]

            subprocess.run(cmd, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            return

    except subprocess.CalledProcessError as e:
        logging.error(f"Sound playback failed: {e}")
    except Exception as e:
        logging.error(f"Unexpected error: {e}")


def on_receive(packet: Dict[str, Any], interface: Any) -> None:
    """
    Handles an incoming packet from a Meshtastic interface.

    Args:
        packet: The received Meshtastic packet as a dictionary.
        interface: The Meshtastic interface instance that received the packet.
    """
    with app_state.lock:
        # Update packet log
        ui_state.packet_buffer.append(packet)
        if len(ui_state.packet_buffer) > 20:
            # Trim buffer to 20 packets
            ui_state.packet_buffer = ui_state.packet_buffer[-20:]

        if ui_state.display_log:
            draw_packetlog_win()

            if ui_state.current_window == 4:
                menu_state.need_redraw = True
        try:
            if "decoded" not in packet:
                return

            # Assume any incoming packet could update the last seen time for a node
            changed = refresh_node_list()
            if changed:
                draw_node_list()

            if packet["decoded"]["portnum"] == "NODEINFO_APP":
                if "user" in packet["decoded"] and "longName" in packet["decoded"]["user"]:
                    maybe_store_nodeinfo_in_db(packet)

            elif packet["decoded"]["portnum"] == "TEXT_MESSAGE_APP":
                hop_start = packet.get('hopStart', 0)
                hop_limit = packet.get('hopLimit', 0)

                hops = hop_start - hop_limit


                if config.notification_sound == "True":
                    schedule_notification_sound()

                message_bytes = packet["decoded"]["payload"]
                message_string = message_bytes.decode("utf-8")

                refresh_channels = False
                refresh_messages = False

                if packet.get("channel"):
                    channel_number = packet["channel"]
                else:
                    channel_number = 0

                if packet["to"] == interface_state.myNodeNum:
                    if packet["from"] in ui_state.channel_list:
                        pass
                    else:
                        ui_state.channel_list.append(packet["from"])
                        if packet["from"] not in ui_state.all_messages:
                            ui_state.all_messages[packet["from"]] = []
                        update_node_info_in_db(packet["from"], chat_archived=False)
                        refresh_channels = True

                    channel_number = ui_state.channel_list.index(packet["from"])

                channel_id = ui_state.channel_list[channel_number]

                if channel_id != ui_state.channel_list[ui_state.selected_channel]:
                    add_notification(channel_number)
                    refresh_channels = True
                else:
                    refresh_messages = True

                # Add received message to the messages list
                message_from_id = packet["from"]
                message_from_string = get_name_from_database(message_from_id, type="short") + ":"

                add_new_message(channel_id, f"{config.message_prefix} [{hops}] {message_from_string} ", message_string)

                if refresh_channels:
                    draw_channel_list()
                if refresh_messages:
                    draw_messages_window(True)

                save_message_to_db(channel_id, message_from_id, message_string)

        except KeyError as e:
            logging.error(f"Error processing packet: {e}")



from argparse import ArgumentParser


def setup_parser() -> ArgumentParser:
    parser = ArgumentParser(
        add_help=True,
        epilog="If no connection arguments are specified, we attempt a serial connection and then a TCP connection to localhost.",
    )

    connOuter = parser.add_argument_group(
        "Connection", "Optional arguments to specify a device to connect to and how."
    )
    conn = connOuter.add_mutually_exclusive_group()
    conn.add_argument(
        "--port",
        "--serial",
        "-s",
        help="The port to connect to via serial, e.g. `/dev/ttyUSB0`.",
        nargs="?",
        default=None,
        const=None,
    )
    conn.add_argument(
        "--host",
        "--tcp",
        "-t",
        help="The hostname or IP address to connect to using TCP.",
        nargs="?",
        default=None,
        const="localhost",
    )
    conn.add_argument(
        "--ble", "-b", help="The BLE device MAC address or name to connect to.", nargs="?", default=None, const="any"
    )
    parser.add_argument(
        "--settings", "--set", "--control", "-c", help="Launch directly into the settings", action="store_true"
    )

    return parser



import logging
import meshtastic.serial_interface, meshtastic.tcp_interface, meshtastic.ble_interface


def initialize_interface(args):
    try:

        if args.ble:
            return meshtastic.ble_interface.BLEInterface(args.ble if args.ble != "any" else None)

        elif args.host:
            try:
                if ":" in args.host:
                    tcp_hostname, tcp_port = args.host.split(":")
                else:
                    tcp_hostname = args.host
                    tcp_port = meshtastic.tcp_interface.DEFAULT_TCP_PORT
                return meshtastic.tcp_interface.TCPInterface(tcp_hostname, portNumber=tcp_port)
            except Exception as ex:
                logging.error(f"Error connecting to {args.host}. {ex}")
        else:
            try:
                client = meshtastic.serial_interface.SerialInterface(args.port)
            except FileNotFoundError as ex:
                logging.error(f"The serial device at '{args.port}' was not found. {ex}")
            except PermissionError as ex:
                logging.error(
                    f"You probably need to add yourself to the `dialout` group to use a serial connection. {ex}"
                )
            except Exception as ex:
                logging.error(f"Unexpected error initializing interface: {ex}")
            except OSError as ex:
                logging.error(f"The serial device couldn't be opened, it might be in use by another process. {ex}")
            if client.devPath is None:
                try:
                    client = meshtastic.tcp_interface.TCPInterface("localhost")
                except Exception as ex:
                    logging.error(f"Error connecting to localhost:{ex}")

            return client

    except Exception as ex:
        logging.critical(f"Fatal error initializing interface: {ex}")



import json
import logging
import os
from typing import Dict, List, Optional
from contact.ui.colors import setup_colors

# Get the parent directory of the script
script_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.abspath(os.path.join(script_dir, os.pardir))

# To test writting to a non-writable directory, you can uncomment the following lines:
# mkdir /tmp/test_nonwritable
# chmod -w /tmp/test_nonwritable
# parent_dir = "/tmp/test_nonwritable"


def reload_config() -> None:
    loaded_config = initialize_config()
    assign_config_variables(loaded_config)
    setup_colors(reinit=True)


def _is_writable_dir(path: str) -> bool:
    """
    Return True if we can create & delete a temp file in `path`.
    """
    if not os.path.isdir(path):
        return False
    test_path = os.path.join(path, ".perm_test_tmp")
    try:
        with open(test_path, "w", encoding="utf-8") as _tmp:
            _tmp.write("ok")
        os.remove(test_path)
        return True
    except OSError:
        return False


def _get_config_root(preferred_dir: str, fallback_name: str = ".contact_client") -> str:
    """
    Choose a writable directory for config artifacts.
    """
    if _is_writable_dir(preferred_dir):
        return preferred_dir

    home = os.path.expanduser("~")
    fallback_dir = os.path.join(home, fallback_name)
    # Ensure the fallback exists.
    os.makedirs(fallback_dir, exist_ok=True)

    # If *that* still isn't writable, last-ditch: use a system temp dir.
    if not _is_writable_dir(fallback_dir):
        import tempfile

        fallback_dir = tempfile.mkdtemp(prefix="contact_client_")

    return fallback_dir


# Pick the root now.
config_root = _get_config_root(parent_dir)

# Paths (derived from the chosen root)
json_file_path = os.path.join(config_root, "config.json")
log_file_path = os.path.join(config_root, "client.log")
db_file_path = os.path.join(config_root, "client.db")
node_configs_file_path = os.path.join(config_root, "node-configs/")
localisations_dir = os.path.join(parent_dir, "localisations")


def get_localisation_options(localisations_path: Optional[str] = None) -> List[str]:
    """
    Return available localisation codes from the localisations folder.
    """
    localisations_path = localisations_path or localisations_dir
    if not os.path.isdir(localisations_path):
        return []

    options = []
    for filename in os.listdir(localisations_path):
        if filename.startswith(".") or not filename.endswith(".ini"):
            continue
        options.append(os.path.splitext(filename)[0])

    return sorted(options)


def get_localisation_file(language: str, localisations_path: Optional[str] = None) -> str:
    """
    Return a valid localisation file path, falling back to a default when missing.
    """
    localisations_path = localisations_path or localisations_dir
    available = get_localisation_options(localisations_path)
    if not available:
        return os.path.join(localisations_path, "en.ini")

    normalized = (language or "").strip().lower()
    if normalized.endswith(".ini"):
        normalized = normalized[:-4]

    if normalized in available:
        return os.path.join(localisations_path, f"{normalized}.ini")

    fallback = "en" if "en" in available else available[0]
    return os.path.join(localisations_path, f"{fallback}.ini")


def format_json_single_line_arrays(data: Dict[str, object], indent: int = 4) -> str:
    """
    Formats JSON with arrays on a single line while keeping other elements properly indented.
    """

    def format_value(value: object, current_indent: int) -> str:
        if isinstance(value, dict):
            items = []
            for key, val in value.items():
                items.append(f'{" " * current_indent}"{key}": {format_value(val, current_indent + indent)}')
            return "{\n" + ",\n".join(items) + f"\n{' ' * (current_indent - indent)}}}"
        elif isinstance(value, list):
            return f"[{', '.join(json.dumps(el, ensure_ascii=False) for el in value)}]"
        else:
            return json.dumps(value, ensure_ascii=False)

    return format_value(data, indent)


# Recursive function to check and update nested dictionaries
def update_dict(default: Dict[str, object], actual: Dict[str, object]) -> bool:
    updated = False
    for key, value in default.items():
        if key not in actual:
            actual[key] = value
            updated = True
        elif isinstance(value, dict):
            # Recursively check nested dictionaries
            updated = update_dict(value, actual[key]) or updated
    return updated


def initialize_config() -> Dict[str, object]:
    COLOR_CONFIG_DARK = {
        "default": ["white", "black"],
        "background": [" ", "black"],
        "splash_logo": ["green", "black"],
        "splash_text": ["white", "black"],
        "input": ["white", "black"],
        "node_list": ["white", "black"],
        "channel_list": ["white", "black"],
        "channel_selected": ["green", "black"],
        "rx_messages": ["cyan", "black"],
        "tx_messages": ["green", "black"],
        "timestamps": ["white", "black"],
        "commands": ["white", "black"],
        "window_frame": ["white", "black"],
        "window_frame_selected": ["green", "black"],
        "log_header": ["blue", "black"],
        "log": ["green", "black"],
        "settings_default": ["white", "black"],
        "settings_sensitive": ["red", "black"],
        "settings_save": ["green", "black"],
        "settings_breadcrumbs": ["white", "black"],
        "settings_warning": ["red", "black"],
        "settings_note": ["green", "black"],
        "node_favorite": ["green", "black"],
        "node_ignored": ["red", "black"],
    }
    COLOR_CONFIG_LIGHT = {
        "default": ["black", "white"],
        "background": [" ", "white"],
        "splash_logo": ["green", "white"],
        "splash_text": ["black", "white"],
        "input": ["black", "white"],
        "node_list": ["black", "white"],
        "channel_list": ["black", "white"],
        "channel_selected": ["green", "white"],
        "rx_messages": ["cyan", "white"],
        "tx_messages": ["green", "white"],
        "timestamps": ["black", "white"],
        "commands": ["black", "white"],
        "window_frame": ["black", "white"],
        "window_frame_selected": ["green", "white"],
        "log_header": ["black", "white"],
        "log": ["blue", "white"],
        "settings_default": ["black", "white"],
        "settings_sensitive": ["red", "white"],
        "settings_save": ["green", "white"],
        "settings_breadcrumbs": ["black", "white"],
        "settings_warning": ["red", "white"],
        "settings_note": ["green", "white"],
        "node_favorite": ["green", "white"],
        "node_ignored": ["red", "white"],
    }
    COLOR_CONFIG_GREEN = {
        "default": ["green", "black"],
        "background": [" ", "black"],
        "splash_logo": ["green", "black"],
        "splash_text": ["green", "black"],
        "input": ["green", "black"],
        "node_list": ["green", "black"],
        "channel_list": ["green", "black"],
        "channel_selected": ["cyan", "black"],
        "rx_messages": ["green", "black"],
        "tx_messages": ["green", "black"],
        "timestamps": ["green", "black"],
        "commands": ["green", "black"],
        "window_frame": ["green", "black"],
        "window_frame_selected": ["cyan", "black"],
        "log_header": ["green", "black"],
        "log": ["green", "black"],
        "settings_default": ["green", "black"],
        "settings_sensitive": ["green", "black"],
        "settings_save": ["green", "black"],
        "settings_breadcrumbs": ["green", "black"],
        "settings_save": ["green", "black"],
        "settings_breadcrumbs": ["green", "black"],
        "settings_warning": ["green", "black"],
        "settings_note": ["green", "black"],
        "node_favorite": ["cyan", "green"],
        "node_ignored": ["red", "black"],
    }
    available_languages = get_localisation_options()
    default_language = "en" if "en" in available_languages else (available_languages[0] if available_languages else "en")
    default_config_variables = {
        "channel_list_16ths": "3",
        "node_list_16ths": "5",
        "single_pane_mode": "False",
        "db_file_path": db_file_path,
        "log_file_path": log_file_path,
        "node_configs_file_path": node_configs_file_path,
        "language": default_language,
        "message_prefix": ">>",
        "sent_message_prefix": ">> Sent",
        "notification_symbol": "*",
        "notification_sound": "True",
        "ack_implicit_str": "[◌]",
        "ack_str": "[✓]",
        "nak_str": "[x]",
        "ack_unknown_str": "[…]",
        "node_sort": "lastHeard",
        "theme": "dark",
        "COLOR_CONFIG_DARK": COLOR_CONFIG_DARK,
        "COLOR_CONFIG_LIGHT": COLOR_CONFIG_LIGHT,
        "COLOR_CONFIG_GREEN": COLOR_CONFIG_GREEN,
    }

    if not os.path.exists(json_file_path):
        with open(json_file_path, "w", encoding="utf-8") as json_file:
            formatted_json = format_json_single_line_arrays(default_config_variables)
            json_file.write(formatted_json)

    # Ensure all default variables exist in the JSON file
    with open(json_file_path, "r", encoding="utf-8") as json_file:
        loaded_config = json.load(json_file)

    # Check and add missing variables
    updated = update_dict(default_config_variables, loaded_config)

    # Update the JSON file if any variables were missing
    if updated:
        formatted_json = format_json_single_line_arrays(loaded_config)
        with open(json_file_path, "w", encoding="utf-8") as json_file:
            json_file.write(formatted_json)
        logging.info(f"JSON file updated with missing default variables and COLOR_CONFIG items.")

    return loaded_config


def assign_config_variables(loaded_config: Dict[str, object]) -> None:
    # Assign values to local variables

    global db_file_path, log_file_path, node_configs_file_path, message_prefix, sent_message_prefix
    global notification_symbol, ack_implicit_str, ack_str, nak_str, ack_unknown_str
    global node_list_16ths, channel_list_16ths, single_pane_mode
    global theme, COLOR_CONFIG, language
    global node_sort, notification_sound

    channel_list_16ths = loaded_config["channel_list_16ths"]
    node_list_16ths = loaded_config["node_list_16ths"]
    single_pane_mode = loaded_config["single_pane_mode"]
    db_file_path = loaded_config["db_file_path"]
    log_file_path = loaded_config["log_file_path"]
    node_configs_file_path = loaded_config.get("node_configs_file_path")
    language = loaded_config["language"]
    message_prefix = loaded_config["message_prefix"]
    sent_message_prefix = loaded_config["sent_message_prefix"]
    notification_symbol = loaded_config["notification_symbol"]
    notification_sound = loaded_config["notification_sound"]
    ack_implicit_str = loaded_config["ack_implicit_str"]
    ack_str = loaded_config["ack_str"]
    nak_str = loaded_config["nak_str"]
    ack_unknown_str = loaded_config["ack_unknown_str"]
    node_sort = loaded_config["node_sort"]
    theme = loaded_config["theme"]
    if theme == "dark":
        COLOR_CONFIG = loaded_config["COLOR_CONFIG_DARK"]
    elif theme == "light":
        COLOR_CONFIG = loaded_config["COLOR_CONFIG_LIGHT"]
    elif theme == "green":
        COLOR_CONFIG = loaded_config["COLOR_CONFIG_GREEN"]


# Call the function when the script is imported
loaded_config = initialize_config()
assign_config_variables(loaded_config)

if __name__ == "__main__":
    logging.basicConfig(
        filename="default_config.log",
        level=logging.INFO,  # DEBUG, INFO, WARNING, ERROR, CRITICAL)
        format="%(asctime)s - %(levelname)s - %(message)s",
    )
    print("\nLoaded Configuration:")
    print(f"Database File Path: {db_file_path}")
    print(f"Log File Path: {log_file_path}")
    print(f"Configs File Path: {node_configs_file_path}")
    print(f"Message Prefix: {message_prefix}")
    print(f"Sent Message Prefix: {sent_message_prefix}")
    print(f"Notification Symbol: {notification_symbol}")
    print(f"ACK Implicit String: {ack_implicit_str}")
    print(f"ACK String: {ack_str}")
    print(f"NAK String: {nak_str}")
    print(f"ACK Unknown String: {ack_unknown_str}")
    print(f"Color Config: {COLOR_CONFIG}")



"""Helpers for normalizing emoji sequences in width-sensitive message rendering."""

# Strip zero-width and presentation modifiers that make terminal cell width inconsistent.
EMOJI_MODIFIER_REPLACEMENTS = {
    "\u200d": "",
    "\u20e3": "",
    "\ufe0e": "",
    "\ufe0f": "",
    "\U0001F3FB": "",
    "\U0001F3FC": "",
    "\U0001F3FD": "",
    "\U0001F3FE": "",
    "\U0001F3FF": "",
}

_EMOJI_MODIFIER_TRANSLATION = str.maketrans(EMOJI_MODIFIER_REPLACEMENTS)
_REGIONAL_INDICATOR_START = ord("\U0001F1E6")
_REGIONAL_INDICATOR_END = ord("\U0001F1FF")


def _regional_indicator_to_letter(char: str) -> str:
    return chr(ord("A") + ord(char) - _REGIONAL_INDICATOR_START)


def _normalize_flag_emoji(text: str) -> str:
    """Convert flag emoji built from regional indicators into ASCII country codes."""
    normalized = []
    index = 0

    while index < len(text):
        current = text[index]
        current_ord = ord(current)

        if _REGIONAL_INDICATOR_START <= current_ord <= _REGIONAL_INDICATOR_END and index + 1 < len(text):
            next_char = text[index + 1]
            next_ord = ord(next_char)
            if _REGIONAL_INDICATOR_START <= next_ord <= _REGIONAL_INDICATOR_END:
                normalized.append(_regional_indicator_to_letter(current))
                normalized.append(_regional_indicator_to_letter(next_char))
                index += 2
                continue

        normalized.append(current)
        index += 1

    return "".join(normalized)


def normalize_message_text(text: str) -> str:
    """Strip modifiers and rewrite flag emoji into stable terminal-friendly text."""
    if not text:
        return text

    return _normalize_flag_emoji(text.translate(_EMOJI_MODIFIER_TRANSLATION))



import os
import json
import curses
from typing import Any, List, Dict, Optional

from contact.ui.colors import get_color, setup_colors, COLOR_MAP
import contact.ui.default_config as config
from contact.ui.nav_utils import move_highlight, draw_arrows, update_help_window
from contact.utilities.ini_utils import parse_ini_file
from contact.utilities.input_handlers import get_list_input
from contact.utilities.i18n import t
from contact.utilities.singleton import menu_state


MAX_MENU_WIDTH = 80  # desired max; will shrink on small terminals
max_help_lines = 6
save_option = "Save Changes"
translation_file = config.get_localisation_file(config.language)
field_mapping, help_text = parse_ini_file(translation_file)
translation_language = config.language


def reload_translations(language: Optional[str] = None) -> None:
    global translation_file, field_mapping, help_text, translation_language
    target_language = language or config.language
    translation_file = config.get_localisation_file(target_language)
    field_mapping, help_text = parse_ini_file(translation_file)
    translation_language = target_language


def get_app_settings_key(menu_path: List[str], selected_key: str) -> str:
    parts = ["app_settings"]
    for part in menu_path:
        if part in ("Main Menu", "App Settings"):
            continue
        parts.append(part)
    parts.append(selected_key)
    return ".".join(parts)


def get_app_settings_path_parts(menu_path: List[str]) -> List[str]:
    parts = ["app_settings"]
    for part in menu_path:
        if part in ("Main Menu", "App Settings"):
            continue
        parts.append(part)
    return parts


def lookup_app_settings_label(full_key: str, fallback: str) -> str:
    label = field_mapping.get(full_key)
    if label:
        return label
    parts = full_key.split(".")
    if len(parts) >= 2 and parts[1].startswith("COLOR_CONFIG_"):
        unified_key = ".".join([parts[0], "color_config"] + parts[2:])
        return field_mapping.get(unified_key, fallback)
    return fallback


def get_app_settings_help_path_parts(menu_path: List[str]) -> List[str]:
    parts = get_app_settings_path_parts(menu_path)
    if parts and parts[-1] in ("COLOR_CONFIG_DARK", "COLOR_CONFIG_LIGHT", "COLOR_CONFIG_GREEN"):
        parts[-1] = "color_config"
    return parts


def get_app_settings_header(menu_path: List[str]) -> str:
    if not menu_path:
        return ""
    translated_parts = []
    for idx, part in enumerate(menu_path):
        if idx == 0:
            translated_parts.append(field_mapping.get(part, part))
            continue
        if part in ("Main Menu", "App Settings"):
            continue
        full_key = ".".join(get_app_settings_path_parts(menu_path[: idx + 1]))
        translated_parts.append(lookup_app_settings_label(full_key, part))
    return " > ".join(translated_parts)


# Compute an effective width that fits the current terminal
def get_effective_width() -> int:
    # Leave space for borders; ensure a sane minimum
    return max(20, min(MAX_MENU_WIDTH, curses.COLS - 2))


def edit_color_pair(key: str, display_label: str, current_value: List[str]) -> List[str]:
    """
    Allows the user to select a foreground and background color for a key.
    """
    color_list = [" "] + list(COLOR_MAP.keys())
    fg_color = get_list_input(
        t(
            "ui.prompt.select_foreground_color",
            default="Select Foreground Color for {label}",
            label=display_label,
        ),
        current_value[0],
        color_list,
    )
    bg_color = get_list_input(
        t(
            "ui.prompt.select_background_color",
            default="Select Background Color for {label}",
            label=display_label,
        ),
        current_value[1],
        color_list,
    )

    return [fg_color, bg_color]


def edit_value(key: str, display_label: str, current_value: str) -> str:

    w = get_effective_width()
    height = 10
    input_width = w - 16  # Allow space for "New Value: "
    start_y = (curses.LINES - height) // 2
    start_x = max(0, (curses.COLS - w) // 2)

    # Create a centered window
    edit_win = curses.newwin(height, w, start_y, start_x)
    edit_win.bkgd(get_color("background"))
    edit_win.attrset(get_color("window_frame"))
    edit_win.border()

    # Display instructions
    edit_win.addstr(
        1,
        2,
        t("ui.label.editing", default="Editing {label}", label=display_label),
        get_color("settings_default", bold=True),
    )
    edit_win.addstr(3, 2, t("ui.label.current_value", default="Current Value:"), get_color("settings_default"))

    wrap_width = w - 4  # Account for border and padding
    wrapped_lines = [current_value[i : i + wrap_width] for i in range(0, len(current_value), wrap_width)]

    for i, line in enumerate(wrapped_lines[:4]):  # Limit display to fit the window height
        edit_win.addstr(4 + i, 2, line, get_color("settings_default"))

    edit_win.refresh()

    # Handle theme selection dynamically
    if key == "theme":
        # Load theme names dynamically from the JSON
        theme_options = [
            k.split("_", 2)[2].lower() for k in config.loaded_config.keys() if k.startswith("COLOR_CONFIG")
        ]
        return get_list_input(
            t("ui.prompt.select_value", default="Select {label}", label=display_label),
            current_value,
            theme_options,
        )

    elif key == "language":
        language_options = config.get_localisation_options()
        if not language_options:
            return current_value
        return get_list_input(
            t("ui.prompt.select_value", default="Select {label}", label=display_label),
            current_value,
            language_options,
        )

    elif key == "node_sort":
        sort_options = ["lastHeard", "name", "hops"]
        return get_list_input(display_label, current_value, sort_options)

    elif key == "notification_sound":
        sound_options = ["True", "False"]
        return get_list_input(display_label, current_value, sound_options)

    elif key == "single_pane_mode":
        sound_options = ["True", "False"]
        return get_list_input(display_label, current_value, sound_options)

    # Standard Input Mode (Scrollable)
    edit_win.addstr(7, 2, t("ui.label.new_value", default="New Value: "), get_color("settings_default"))
    curses.curs_set(1)

    scroll_offset = 0  # Determines which part of the text is visible
    user_input = ""
    input_position = (7, 13)  # Tuple for row and column
    row, col = input_position  # Unpack tuple

    while True:
        if menu_state.need_redraw:
            curses.update_lines_cols()
            menu_state.need_redraw = False

            # Re-create the window to fully reset state
            edit_win = curses.newwin(height, w, start_y, start_x)
            edit_win.timeout(200)
            edit_win.bkgd(get_color("background"))
            edit_win.attrset(get_color("window_frame"))
            edit_win.border()

            # Redraw static content
            edit_win.addstr(
                1,
                2,
                t("ui.label.editing", default="Editing {label}", label=display_label),
                get_color("settings_default", bold=True),
            )
            edit_win.addstr(
                3, 2, t("ui.label.current_value", default="Current Value:"), get_color("settings_default")
            )
            for i, line in enumerate(wrapped_lines[:4]):
                edit_win.addstr(4 + i, 2, line, get_color("settings_default"))
            edit_win.addstr(
                7, 2, t("ui.label.new_value", default="New Value: "), get_color("settings_default")
            )

        visible_text = user_input[scroll_offset : scroll_offset + input_width]
        edit_win.addstr(row, col, " " * input_width, get_color("settings_default"))
        edit_win.addstr(row, col, visible_text, get_color("settings_default"))
        edit_win.refresh()

        edit_win.move(row, col + min(len(user_input) - scroll_offset, input_width))

        try:
            key = edit_win.get_wch()
        except curses.error:
            continue  # window not ready — skip this loop

        if key in (chr(27), curses.KEY_LEFT):
            curses.curs_set(0)
            return current_value

        elif key in (chr(curses.KEY_ENTER), chr(10), chr(13)):
            break

        elif key in (curses.KEY_BACKSPACE, chr(127)):
            if user_input:
                user_input = user_input[:-1]
                if scroll_offset > 0 and len(user_input) < scroll_offset + input_width:
                    scroll_offset -= 1
        else:
            if isinstance(key, str):
                user_input += key
            else:
                user_input += chr(key)

            if len(user_input) > input_width:
                scroll_offset += 1

    curses.curs_set(0)
    return user_input if user_input else current_value


def display_menu() -> tuple[Any, Any, List[str]]:
    """
    Render the configuration menu with a Save button directly added to the window.
    """
    if translation_language != config.language:
        reload_translations()

    num_items = len(menu_state.current_menu) + (1 if menu_state.show_save_option else 0)

    # Determine menu items based on the type of current_menu
    if isinstance(menu_state.current_menu, dict):
        options = list(menu_state.current_menu.keys())
    elif isinstance(menu_state.current_menu, list):
        options = [f"[{i}]" for i in range(len(menu_state.current_menu))]
    else:
        options = []  # Fallback in case of unexpected data types

    # Calculate dynamic dimensions for the menu
    min_help_window_height = 6
    max_menu_height = curses.LINES
    menu_height = min(max_menu_height - min_help_window_height, num_items + 5)
    num_items = len(options)
    w = get_effective_width()
    start_y = (curses.LINES - menu_height) // 2 - (min_help_window_height // 2)
    start_x = max(0, (curses.COLS - w) // 2)

    # Create the window
    menu_win = curses.newwin(menu_height, w, start_y, start_x)
    menu_win.erase()
    menu_win.bkgd(get_color("background"))
    menu_win.attrset(get_color("window_frame"))
    menu_win.border()
    menu_win.keypad(True)

    # Create the pad for scrolling
    menu_pad = curses.newpad(num_items + 1, w - 8)
    menu_pad.bkgd(get_color("background"))

    # Display the menu path
    header = get_app_settings_header(menu_state.menu_path)
    if len(header) > w - 4:
        header = header[: w - 7] + "..."
    menu_win.addstr(1, 2, header, get_color("settings_breadcrumbs", bold=True))

    # Populate the pad with menu options
    for idx, key in enumerate(options):
        value = (
            menu_state.current_menu[key]
            if isinstance(menu_state.current_menu, dict)
            else menu_state.current_menu[int(key.strip("[]"))]
        )
        if isinstance(menu_state.current_menu, dict):
            full_key = get_app_settings_key(menu_state.menu_path, key)
            display_key = lookup_app_settings_label(full_key, key)
        else:
            display_key = key
        display_key = f"{display_key}"[: w // 2 - 2]
        display_value = f"{value}"[: w // 2 - 8]

        color = get_color("settings_default", reverse=(idx == menu_state.selected_index))
        menu_pad.addstr(idx, 0, f"{display_key:<{w // 2 - 2}} {display_value}".ljust(w - 8), color)

    # Add Save button to the main window
    if menu_state.show_save_option:
        save_position = menu_height - 2
        save_label = t("ui.save_changes", default=save_option)
        menu_win.addstr(
            save_position,
            (w - len(save_label)) // 2,
            save_label,
            get_color("settings_save", reverse=(menu_state.selected_index == len(menu_state.current_menu))),
        )

    menu_win.refresh()
    menu_pad.refresh(
        menu_state.start_index[-1],
        0,
        menu_win.getbegyx()[0] + 3,
        menu_win.getbegyx()[1] + 4,
        menu_win.getbegyx()[0] + 3 + menu_win.getmaxyx()[0] - 5 - (2 if menu_state.show_save_option else 0),
        menu_win.getbegyx()[1] + menu_win.getmaxyx()[1] - 4,
    )

    max_index = num_items + (1 if menu_state.show_save_option else 0) - 1
    visible_height = menu_win.getmaxyx()[0] - 5 - (2 if menu_state.show_save_option else 0)

    draw_arrows(menu_win, visible_height, max_index, menu_state.start_index, menu_state.show_save_option)

    # Draw help window below the menu
    global max_help_lines
    remaining_space = curses.LINES - (start_y + menu_height + 2)
    max_help_lines = max(remaining_space, 1)
    transformed_path = get_app_settings_help_path_parts(menu_state.menu_path)
    selected_option = (
        options[min(menu_state.selected_index, len(options) - 1)] if options and menu_state.selected_index >= 0 else None
    )
    help_y = menu_win.getbegyx()[0] + menu_win.getmaxyx()[0]
    menu_state.help_win = update_help_window(
        menu_state.help_win,
        help_text,
        transformed_path,
        selected_option,
        max_help_lines,
        w,
        help_y,
        menu_win.getbegyx()[1],
    )

    return menu_win, menu_pad, options


def update_app_settings_help(menu_win: curses.window, options: List[str]) -> None:
    transformed_path = get_app_settings_help_path_parts(menu_state.menu_path)
    selected_option = options[menu_state.selected_index] if menu_state.selected_index < len(options) else None
    help_y = menu_win.getbegyx()[0] + menu_win.getmaxyx()[0]
    menu_state.help_win = update_help_window(
        menu_state.help_win,
        help_text,
        transformed_path,
        selected_option,
        max_help_lines,
        menu_win.getmaxyx()[1],
        help_y,
        menu_win.getbegyx()[1],
    )


def json_editor(stdscr: curses.window, menu_state: Any) -> None:

    menu_state.selected_index = 0  # Track the selected option
    made_changes = False  # Track if any changes were made

    script_dir = os.path.dirname(os.path.abspath(__file__))
    parent_dir = os.path.abspath(os.path.join(script_dir, os.pardir))
    file_path = os.path.join(parent_dir, "config.json")

    menu_state.show_save_option = True  # Always show the Save button
    menu_state.help_win = None
    menu_state.help_text = {}

    # Ensure the file exists
    if not os.path.exists(file_path):
        with open(file_path, "w") as f:
            json.dump({}, f)

    # Load JSON data
    with open(file_path, "r", encoding="utf-8") as f:
        original_data = json.load(f)

    data = original_data  # Reference to the original data
    menu_state.current_menu = data  # Track the current level of the menu

    # Render the menu
    menu_win, menu_pad, options = display_menu()
    update_app_settings_help(menu_win, options)
    menu_state.need_redraw = True

    while True:
        if menu_state.need_redraw:
            menu_state.need_redraw = False
            menu_win, menu_pad, options = display_menu()
            menu_win.refresh()
            update_app_settings_help(menu_win, options)

        max_index = len(options) + (1 if menu_state.show_save_option else 0) - 1

        menu_win.timeout(200)
        key = menu_win.getch()

        if key == curses.KEY_UP:

            old_selected_index = menu_state.selected_index
            menu_state.selected_index = max_index if menu_state.selected_index == 0 else menu_state.selected_index - 1
            menu_state.help_win = move_highlight(
                old_selected_index, options, menu_win, menu_pad, menu_state=menu_state, max_help_lines=max_help_lines
            )
            update_app_settings_help(menu_win, options)

        elif key == curses.KEY_DOWN:

            old_selected_index = menu_state.selected_index
            menu_state.selected_index = 0 if menu_state.selected_index == max_index else menu_state.selected_index + 1
            menu_state.help_win = move_highlight(
                old_selected_index, options, menu_win, menu_pad, menu_state=menu_state, max_help_lines=max_help_lines
            )
            update_app_settings_help(menu_win, options)

        elif key == ord("\t") and menu_state.show_save_option:
            old_selected_index = menu_state.selected_index
            menu_state.selected_index = max_index
            menu_state.help_win = move_highlight(
                old_selected_index, options, menu_win, menu_pad, menu_state=menu_state, max_help_lines=max_help_lines
            )
            update_app_settings_help(menu_win, options)

        elif key in (curses.KEY_RIGHT, 10, 13):  # 10 = \n, 13 = carriage return

            menu_state.need_redraw = True
            menu_win.erase()
            menu_win.refresh()
            if menu_state.help_win:
                menu_state.help_win.erase()
                menu_state.help_win.refresh()

            if menu_state.selected_index < len(options):  # Handle selection of a menu item
                selected_key = options[menu_state.selected_index]
                menu_state.menu_path.append(str(selected_key))
                menu_state.start_index.append(0)
                menu_state.menu_index.append(menu_state.selected_index)

                # Handle nested data
                if isinstance(menu_state.current_menu, dict):
                    if selected_key in menu_state.current_menu:
                        selected_data = menu_state.current_menu[selected_key]
                    else:
                        continue  # Skip invalid key
                elif isinstance(menu_state.current_menu, list):
                    selected_data = menu_state.current_menu[int(selected_key.strip("[]"))]

                display_label = selected_key
                if isinstance(menu_state.current_menu, dict):
                    path_for_label = (
                        menu_state.menu_path[:-1]
                        if menu_state.menu_path and menu_state.menu_path[-1] == str(selected_key)
                        else menu_state.menu_path
                    )
                    full_key = get_app_settings_key(path_for_label, selected_key)
                    display_label = lookup_app_settings_label(full_key, selected_key)

                if isinstance(selected_data, list) and len(selected_data) == 2:
                    # Edit color pair
                    old = selected_data
                    new_value = edit_color_pair(selected_key, display_label, selected_data)
                    menu_state.menu_path.pop()
                    menu_state.start_index.pop()
                    menu_state.menu_index.pop()
                    menu_state.current_menu[selected_key] = new_value
                    if new_value != old:
                        made_changes = True

                elif isinstance(selected_data, (dict, list)):
                    # Navigate into nested data
                    menu_state.current_menu = selected_data
                    menu_state.selected_index = 0  # Reset the selected index

                else:
                    # General value editing
                    old = selected_data
                    new_value = edit_value(selected_key, display_label, selected_data)
                    menu_state.menu_path.pop()
                    menu_state.menu_index.pop()
                    menu_state.start_index.pop()
                    menu_state.current_menu[selected_key] = new_value
                    menu_state.need_redraw = True
                    if new_value != old:
                        made_changes = True

            else:
                # Save button selected
                save_json(file_path, data)
                made_changes = False
                stdscr.refresh()
                # config.reload()  # This isn't refreshing the file paths as expected
                break

        elif key in (27, curses.KEY_LEFT):  # Escape or Left Arrow

            menu_state.need_redraw = True
            menu_win.erase()
            menu_win.refresh()
            if menu_state.help_win:
                menu_state.help_win.erase()
                menu_state.help_win.refresh()

            # menu_state.selected_index = menu_state.menu_index[-1]

            # Navigate back in the menu
            if len(menu_state.menu_path) > 2:
                menu_state.menu_path.pop()
                menu_state.start_index.pop()
                menu_state.current_menu = data

                for path in menu_state.menu_path[2:]:
                    menu_state.current_menu = (
                        menu_state.current_menu[path]
                        if isinstance(menu_state.current_menu, dict)
                        else menu_state.current_menu[int(path.strip("[]"))]
                    )

            else:
                # Exit the editor
                if made_changes:
                    save_prompt = get_list_input(
                        t("ui.confirm.save_before_exit", default="You have unsaved changes. Save before exiting?"),
                        None,
                        ["Yes", "No", "Cancel"],
                        mandatory=True,
                    )
                    if save_prompt == "Cancel":
                        continue  # Stay in the menu without doing anything
                    elif save_prompt == "Yes":
                        save_json(file_path, data)
                        made_changes = False

                menu_win.clear()
                menu_win.refresh()

                break


def save_json(file_path: str, data: Dict[str, Any]) -> None:
    formatted_json = config.format_json_single_line_arrays(data)
    with open(file_path, "w", encoding="utf-8") as f:
        f.write(formatted_json)
    config.reload_config()
    reload_translations(data.get("language"))


def main(stdscr: curses.window) -> None:
    from contact.ui.ui_state import MenuState

    if len(menu_state.menu_path) == 0:
        menu_state.menu_path = ["App Settings"]  # Initialize if not set

    curses.curs_set(0)
    stdscr.keypad(True)
    setup_colors()
    json_editor(stdscr, menu_state)


if __name__ == "__main__":
    curses.wrapper(main)



import base64
import curses
import ipaddress
import logging
import os
import sys
from typing import List

from contact.utilities.save_to_radio import save_changes
import contact.ui.default_config as config
from contact.utilities.config_io import config_export, config_import
from contact.utilities.control_utils import transform_menu_path
from contact.utilities.i18n import t
from contact.utilities.ini_utils import parse_ini_file
from contact.utilities.input_handlers import (
    get_repeated_input,
    get_text_input,
    get_fixed32_input,
    get_list_input,
    get_admin_key_input,
)
from contact.ui.colors import get_color
from contact.ui.dialog import dialog
from contact.ui.menus import generate_menu_from_protobuf
from contact.ui.nav_utils import move_highlight, draw_arrows, update_help_window
from contact.ui.user_config import json_editor
from contact.utilities.singleton import menu_state

# Setup Variables
MAX_MENU_WIDTH = 80  # desired max; will shrink on small terminals
save_option = "Save Changes"
max_help_lines = 0
help_win = None
sensitive_settings = ["Reboot", "Reset Node DB", "Shutdown", "Factory Reset"]


# Compute the effective menu width for the current terminal
def get_menu_width() -> int:
    # Leave at least 2 columns for borders; clamp to >= 20 for usability
    return max(20, min(MAX_MENU_WIDTH, curses.COLS - 2))


sensitive_settings = ["Reboot", "Reset Node DB", "Shutdown", "Factory Reset"]

# Get the parent directory of the script
script_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.abspath(os.path.join(script_dir, os.pardir))

# Paths
# locals_dir = os.path.dirname(os.path.abspath(sys.argv[0]))  # Current script directory
translation_file = config.get_localisation_file(config.language)

# config_folder = os.path.join(locals_dir, "node-configs")
config_folder = os.path.abspath(config.node_configs_file_path)

# Load translations
field_mapping, help_text = parse_ini_file(translation_file)

def _is_repeated_field(field_desc) -> bool:
    """Return True if the protobuf field is repeated. 
       Protobuf 6.31.0 and later use an is_repeated property, while older versions compare against the label field.
       """
    if hasattr(field_desc, "is_repeated"):
        return bool(field_desc.is_repeated)
    return field_desc.label == field_desc.LABEL_REPEATED

def reload_translations() -> None:
    global translation_file, field_mapping, help_text
    translation_file = config.get_localisation_file(config.language)
    field_mapping, help_text = parse_ini_file(translation_file)


def get_translated_header(menu_path: List[str]) -> str:
    if not menu_path:
        return ""

    transformed_path = transform_menu_path(menu_path)
    translated_parts = []
    for idx, part in enumerate(menu_path):
        if idx == 0:
            translated_parts.append(field_mapping.get(part, part))
            continue
        full_key = ".".join(transformed_path[:idx])
        translated_parts.append(field_mapping.get(full_key, part))
    return " > ".join(translated_parts)


def display_menu() -> tuple[object, object]:
    # if help_win:
    #     min_help_window_height = 6
    # else:
    #     min_help_window_height = 0

    min_help_window_height = 6

    num_items = len(menu_state.current_menu) + (1 if menu_state.show_save_option else 0)

    # Determine the available height for the menu
    max_menu_height = curses.LINES
    menu_height = min(max_menu_height - min_help_window_height, num_items + 5)
    w = get_menu_width()
    start_y = (curses.LINES - menu_height) // 2 - (min_help_window_height // 2)
    start_x = (curses.COLS - w) // 2

    # Calculate remaining space for help window
    global max_help_lines
    remaining_space = curses.LINES - (start_y + menu_height + 2)  # +2 for padding
    max_help_lines = max(remaining_space, 1)  # Ensure at least 1 lines for help

    menu_win = curses.newwin(menu_height, w, start_y, start_x)
    menu_win.erase()
    menu_win.bkgd(get_color("background"))
    menu_win.attrset(get_color("window_frame"))
    menu_win.border()
    menu_win.keypad(True)

    menu_pad = curses.newpad(len(menu_state.current_menu) + 1, w - 8)
    menu_pad.bkgd(get_color("background"))

    header = get_translated_header(menu_state.menu_path)
    if len(header) > w - 4:
        header = header[: w - 7] + "..."
    menu_win.addstr(1, 2, header, get_color("settings_breadcrumbs", bold=True))

    transformed_path = transform_menu_path(menu_state.menu_path)

    for idx, option in enumerate(menu_state.current_menu):
        field_info = menu_state.current_menu[option]
        current_value = field_info[1] if isinstance(field_info, tuple) else ""
        full_key = ".".join(transformed_path + [option])
        display_name = field_mapping.get(full_key, option)

        if full_key.startswith("config.network.ipv4_config.") and option in {"ip", "gateway", "subnet", "dns"}:
            if isinstance(current_value, int):
                try:
                    current_value = str(
                        ipaddress.IPv4Address(int(current_value).to_bytes(4, "little", signed=False))
                    )
                except ipaddress.AddressValueError:
                    pass
            elif isinstance(current_value, str) and current_value.isdigit():
                try:
                    current_value = str(
                        ipaddress.IPv4Address(int(current_value).to_bytes(4, "little", signed=False))
                    )
                except ipaddress.AddressValueError:
                    pass

        display_option = f"{display_name}"[: w // 2 - 2]
        display_value = f"{current_value}"[: w // 2 - 4]

        try:
            color = get_color(
                "settings_sensitive" if option in sensitive_settings else "settings_default",
                reverse=(idx == menu_state.selected_index),
            )
            menu_pad.addstr(idx, 0, f"{display_option:<{w // 2 - 2}} {display_value}".ljust(w - 8), color)
        except curses.error:
            pass

    if menu_state.show_save_option:
        save_position = menu_height - 2
        save_label = t("ui.save_changes", default=save_option)
        menu_win.addstr(
            save_position,
            (w - len(save_label)) // 2,
            save_label,
            get_color("settings_save", reverse=(menu_state.selected_index == len(menu_state.current_menu))),
        )

    # Draw help window with dynamically updated max_help_lines
    draw_help_window(start_y, start_x, menu_height, max_help_lines, transformed_path)

    menu_win.refresh()
    menu_pad.refresh(
        menu_state.start_index[-1],
        0,
        menu_win.getbegyx()[0] + 3,
        menu_win.getbegyx()[1] + 4,
        menu_win.getbegyx()[0] + 3 + menu_win.getmaxyx()[0] - 5 - (2 if menu_state.show_save_option else 0),
        menu_win.getbegyx()[1] + menu_win.getmaxyx()[1] - 4,
    )
    curses.curs_set(0)

    max_index = num_items + (1 if menu_state.show_save_option else 0) - 1
    visible_height = menu_win.getmaxyx()[0] - 5 - (2 if menu_state.show_save_option else 0)

    draw_arrows(menu_win, visible_height, max_index, menu_state.start_index, menu_state.show_save_option)

    return menu_win, menu_pad


def draw_help_window(
    menu_start_y: int,
    menu_start_x: int,
    menu_height: int,
    max_help_lines: int,
    transformed_path: List[str],
) -> None:
    global help_win

    if "help_win" not in globals():
        help_win = None  # Initialize if it does not exist

    selected_option = (
        list(menu_state.current_menu.keys())[menu_state.selected_index] if menu_state.current_menu else None
    )
    help_y = menu_start_y + menu_height

    # Use current terminal width for the help window width calculation
    help_win = update_help_window(
        help_win, help_text, transformed_path, selected_option, max_help_lines, get_menu_width(), help_y, menu_start_x
    )


def get_input_type_for_field(field) -> type:
    if field.type in (field.TYPE_INT32, field.TYPE_UINT32, field.TYPE_INT64):
        return int
    elif field.type in (field.TYPE_FLOAT, field.TYPE_DOUBLE):
        return float
    else:
        return str


def settings_menu(stdscr: object, interface: object) -> None:
    curses.update_lines_cols()

    menu = generate_menu_from_protobuf(interface)
    menu_state.current_menu = menu["Main Menu"]
    menu_state.menu_path = ["Main Menu"]

    modified_settings = {}

    menu_state.need_redraw = True
    menu_state.show_save_option = False
    new_value_name = None

    while True:
        if menu_state.need_redraw:
            menu_state.need_redraw = False
            options = list(menu_state.current_menu.keys())

            # Determine if save option should be shown
            path = menu_state.menu_path
            menu_state.show_save_option = (
                (len(path) > 2 and ("Radio Settings" in path or "Module Settings" in path))
                or (len(path) == 2 and "User Settings" in path)
                or (len(path) == 3 and "Channels" in path)
            )

            # Display the menu
            menu_win, menu_pad = display_menu()

        if menu_win is None:
            continue  # Skip if menu_win is not initialized

        menu_win.timeout(200)  # wait up to 200 ms for a keypress (or less if key is pressed)
        key = menu_win.getch()
        if key == -1:
            continue

        max_index = len(options) + (1 if menu_state.show_save_option else 0) - 1
        # max_help_lines = 4

        if key == curses.KEY_UP:
            old_selected_index = menu_state.selected_index
            menu_state.selected_index = max_index if menu_state.selected_index == 0 else menu_state.selected_index - 1
            move_highlight(
                old_selected_index,
                options,
                menu_win,
                menu_pad,
                menu_state=menu_state,
                help_win=help_win,
                help_text=help_text,
                max_help_lines=max_help_lines,
            )

        elif key == curses.KEY_DOWN:
            old_selected_index = menu_state.selected_index
            menu_state.selected_index = 0 if menu_state.selected_index == max_index else menu_state.selected_index + 1
            move_highlight(
                old_selected_index,
                options,
                menu_win,
                menu_pad,
                menu_state=menu_state,
                help_win=help_win,
                help_text=help_text,
                max_help_lines=max_help_lines,
            )

        elif key == curses.KEY_RESIZE:
            menu_state.need_redraw = True
            curses.update_lines_cols()

            menu_win.erase()
            if help_win:
                help_win.erase()

            menu_win.refresh()
            if help_win:
                help_win.refresh()

        elif key == ord("\t") and menu_state.show_save_option:
            old_selected_index = menu_state.selected_index
            menu_state.selected_index = max_index
            move_highlight(
                old_selected_index,
                options,
                menu_win,
                menu_pad,
                menu_state=menu_state,
                help_win=help_win,
                help_text=help_text,
                max_help_lines=max_help_lines,
            )

        elif key == curses.KEY_RIGHT or key == ord("\n"):
            menu_state.need_redraw = True
            menu_state.start_index.append(0)
            menu_win.erase()
            if help_win:
                help_win.erase()

            # draw_help_window(menu_win.getbegyx()[0], menu_win.getbegyx()[1], menu_win.getmaxyx()[0], max_help_lines, menu_state.current_menu, selected_index, transform_menu_path(menu_state.menu_path))

            menu_win.refresh()
            if help_win:
                help_win.refresh()

            if menu_state.show_save_option and menu_state.selected_index == len(options):
                save_changes(interface, modified_settings, menu_state)
                modified_settings.clear()
                logging.info("Changes Saved")

                if len(menu_state.menu_path) > 1:
                    menu_state.menu_path.pop()
                    menu_state.current_menu = menu["Main Menu"]
                    for step in menu_state.menu_path[1:]:
                        menu_state.current_menu = menu_state.current_menu.get(step, {})
                    menu_state.selected_index = 0
                continue

            selected_option = options[menu_state.selected_index]

            if selected_option == "Exit":
                break

            elif selected_option == "Export Config File":

                filename = get_text_input(
                    t("ui.prompt.config_filename", default="Enter a filename for the config file"),
                    None,
                    None,
                )
                if not filename:
                    logging.info("Export aborted: No filename provided.")
                    menu_state.start_index.pop()
                    continue  # Go back to the menu
                if not filename.lower().endswith(".yaml"):
                    filename += ".yaml"

                try:
                    config_text = config_export(interface)
                    yaml_file_path = os.path.join(config_folder, filename)

                    if os.path.exists(yaml_file_path):
                        overwrite = get_list_input(
                            t(
                                "ui.confirm.overwrite_file",
                                default="{filename} already exists. Overwrite?",
                                filename=filename,
                            ),
                            None,
                            ["Yes", "No"],
                        )
                        if overwrite == "No":
                            logging.info("Export cancelled: User chose not to overwrite.")
                            menu_state.start_index.pop()
                            continue  # Return to menu
                    os.makedirs(os.path.dirname(yaml_file_path), exist_ok=True)
                    with open(yaml_file_path, "w", encoding="utf-8") as file:
                        file.write(config_text)
                    logging.info(f"Config file saved to {yaml_file_path}")
                    dialog(t("ui.dialog.config_saved_title", default="Config File Saved:"), yaml_file_path)
                    menu_state.need_redraw = True
                    menu_state.start_index.pop()
                    continue
                except PermissionError:
                    logging.error(f"Permission denied: Unable to write to {yaml_file_path}")
                except OSError as e:
                    logging.error(f"OS error while saving config: {e}")
                except Exception as e:
                    logging.error(f"Unexpected error: {e}")
                menu_state.start_index.pop()
                continue

            elif selected_option == "Load Config File":

                # Check if folder exists and is not empty
                if not os.path.exists(config_folder) or not any(os.listdir(config_folder)):
                    dialog("", t("ui.dialog.no_config_files", default=" No config files found. Export a config first."))
                    menu_state.need_redraw = True
                    continue  # Return to menu

                file_list = [f for f in os.listdir(config_folder) if os.path.isfile(os.path.join(config_folder, f))]

                # Ensure file_list is not empty before proceeding
                if not file_list:
                    dialog("", t("ui.dialog.no_config_files", default=" No config files found. Export a config first."))
                    menu_state.need_redraw = True
                    continue

                filename = get_list_input(
                    t("ui.prompt.choose_config_file", default="Choose a config file"), None, file_list
                )
                if filename:
                    file_path = os.path.join(config_folder, filename)
                    overwrite = get_list_input(
                        t(
                            "ui.confirm.load_config_file",
                            default="Are you sure you want to load {filename}?",
                            filename=filename,
                        ),
                        None,
                        ["Yes", "No"],
                    )
                    if overwrite == "Yes":
                        config_import(interface, file_path)
                menu_state.start_index.pop()
                continue

            elif selected_option == "Config URL":
                current_value = interface.localNode.getURL()
                new_value = get_text_input(
                    t(
                        "ui.prompt.config_url_current",
                        default="Config URL is currently: {value}",
                        value=current_value,
                    ),
                    None,
                    str,
                )
                if new_value is not None:
                    current_value = new_value
                    overwrite = get_list_input(
                        t("ui.confirm.load_config_url", default="Are you sure you want to load this config?"),
                        None,
                        ["Yes", "No"],
                    )
                    if overwrite == "Yes":
                        interface.localNode.setURL(new_value)
                        logging.info(f"New Config URL sent to node")
                menu_state.start_index.pop()
                continue

            elif selected_option == "Reboot":
                confirmation = get_list_input(
                    t("ui.confirm.reboot", default="Are you sure you want to Reboot?"), None, ["Yes", "No"]
                )
                if confirmation == "Yes":
                    interface.localNode.reboot()
                    logging.info(f"Node Reboot Requested by menu")
                menu_state.start_index.pop()
                continue

            elif selected_option == "Reset Node DB":
                confirmation = get_list_input(
                    t("ui.confirm.reset_node_db", default="Are you sure you want to Reset Node DB?"),
                    None,
                    ["Yes", "No"],
                )
                if confirmation == "Yes":
                    interface.localNode.resetNodeDb()
                    logging.info(f"Node DB Reset Requested by menu")
                menu_state.start_index.pop()
                continue

            elif selected_option == "Shutdown":
                confirmation = get_list_input(
                    t("ui.confirm.shutdown", default="Are you sure you want to Shutdown?"), None, ["Yes", "No"]
                )
                if confirmation == "Yes":
                    interface.localNode.shutdown()
                    logging.info(f"Node Shutdown Requested by menu")
                menu_state.start_index.pop()
                continue

            elif selected_option == "Factory Reset":
                confirmation = get_list_input(
                    t("ui.confirm.factory_reset", default="Are you sure you want to Factory Reset?"),
                    None,
                    ["Yes", "No"],
                )
                if confirmation == "Yes":
                    interface.localNode.factoryReset()
                    logging.info(f"Factory Reset Requested by menu")
                menu_state.start_index.pop()
                continue

            elif selected_option == "App Settings":
                menu_win.clear()
                menu_win.refresh()
                menu_state.menu_path.append("App Settings")
                menu_state.menu_index.append(menu_state.selected_index)
                json_editor(stdscr, menu_state)  # Open the App Settings menu
                reload_translations()
                menu_state.current_menu = menu["Main Menu"]
                menu_state.menu_path = ["Main Menu"]
                menu_state.start_index.pop()
                menu_state.selected_index = 4
                continue

            field_info = menu_state.current_menu.get(selected_option)
            if isinstance(field_info, tuple):
                field, current_value = field_info

                # Transform the menu path to get the full key
                transformed_path = transform_menu_path(menu_state.menu_path)
                full_key = ".".join(transformed_path + [selected_option])

                # Fetch human-readable name from field_mapping
                human_readable_name = field_mapping.get(full_key, selected_option)

                if selected_option in ["longName", "shortName", "isLicensed"]:
                    if selected_option in ["longName", "shortName"]:
                        new_value = get_text_input(
                            f"{human_readable_name} is currently: {current_value}", selected_option, None
                        )
                        new_value = current_value if new_value is None else new_value
                        menu_state.current_menu[selected_option] = (field, new_value)

                    elif selected_option == "isLicensed":
                        new_value = get_list_input(
                            f"{human_readable_name} is currently: {current_value}",
                            str(current_value),
                            ["True", "False"],
                        )
                        new_value = new_value == "True"
                        menu_state.current_menu[selected_option] = (field, new_value)

                    for option, (field, value) in menu_state.current_menu.items():
                        modified_settings[option] = value

                    menu_state.start_index.pop()

                elif selected_option in ["latitude", "longitude", "altitude"]:
                    new_value = get_text_input(
                        f"{human_readable_name} is currently: {current_value}", selected_option, float
                    )
                    new_value = current_value if new_value is None else new_value
                    menu_state.current_menu[selected_option] = (field, new_value)

                    for option in ["latitude", "longitude", "altitude"]:
                        if option in menu_state.current_menu:
                            modified_settings[option] = menu_state.current_menu[option][1]

                    menu_state.start_index.pop()

                elif selected_option == "admin_key":
                    new_values = get_admin_key_input(current_value)
                    new_value = current_value if new_values is None else [base64.b64decode(key) for key in new_values]
                    menu_state.start_index.pop()

                elif field.type == 8:  # Handle boolean type
                    new_value = get_list_input(human_readable_name, str(current_value), ["True", "False"])
                    if new_value == "Not Set":
                        pass  # Leave it as-is
                    else:
                        new_value = new_value == "True" or new_value is True
                    menu_state.start_index.pop()

                elif _is_repeated_field(field):  # Handle repeated field - Not currently used
                    new_value = get_repeated_input(current_value)
                    new_value = current_value if new_value is None else new_value.split(", ")
                    menu_state.start_index.pop()

                elif field.enum_type:  # Enum field
                    enum_options = {v.name: v.number for v in field.enum_type.values}
                    new_value_name = get_list_input(human_readable_name, current_value, list(enum_options.keys()))
                    new_value = enum_options.get(new_value_name, current_value)
                    menu_state.start_index.pop()

                elif field.type == 7:  # Field type 7 corresponds to FIXED32
                    new_value = get_fixed32_input(current_value)
                    menu_state.start_index.pop()

                elif field.type == 13:  # Field type 13 corresponds to UINT32
                    input_type = get_input_type_for_field(field)
                    new_value = get_text_input(
                        f"{human_readable_name} is currently: {current_value}", selected_option, input_type
                    )
                    new_value = current_value if new_value is None else int(new_value)
                    menu_state.start_index.pop()

                elif field.type == 2:  # Field type 13 corresponds to INT64
                    input_type = get_input_type_for_field(field)
                    new_value = get_text_input(
                        f"{human_readable_name} is currently: {current_value}", selected_option, input_type
                    )
                    new_value = current_value if new_value is None else float(new_value)
                    menu_state.start_index.pop()

                else:  # Handle other field types
                    input_type = get_input_type_for_field(field)
                    new_value = get_text_input(
                        f"{human_readable_name} is currently: {current_value}", selected_option, input_type
                    )
                    new_value = current_value if new_value is None else new_value
                    menu_state.start_index.pop()

                for key in menu_state.menu_path[3:]:  # Skip "Main Menu"
                    modified_settings = modified_settings.setdefault(key, {})

                # For comparison, normalize enum numbers to names
                compare_value = new_value
                if field and field.enum_type and isinstance(new_value, int):
                    enum_value_descriptor = field.enum_type.values_by_number.get(new_value)
                    if enum_value_descriptor:
                        compare_value = enum_value_descriptor.name

                if compare_value != current_value:
                    # Save the raw protobuf number, not the name
                    modified_settings[selected_option] = new_value

                # Convert enum string to int
                if field and field.enum_type:
                    enum_value_descriptor = field.enum_type.values_by_number.get(new_value)
                    new_value = enum_value_descriptor.name if enum_value_descriptor else new_value

                menu_state.current_menu[selected_option] = (field, new_value)
            else:
                menu_state.current_menu = menu_state.current_menu[selected_option]
                menu_state.menu_path.append(selected_option)
                menu_state.menu_index.append(menu_state.selected_index)
                menu_state.selected_index = 0

        elif key == curses.KEY_LEFT:

            # If we are at the main menu and there are unsaved changes, prompt to save
            if len(menu_state.menu_path) == 3 and modified_settings:

                current_section = menu_state.menu_path[-1]
                save_prompt = get_list_input(
                    t(
                        "ui.confirm.save_before_exit_section",
                        default="You have unsaved changes in {section}. Save before exiting?",
                        section=current_section,
                    ),
                    None,
                    ["Yes", "No", "Cancel"],
                    mandatory=True,
                )
                if save_prompt == "Cancel":
                    continue  # Stay in the menu without doing anything
                elif save_prompt == "Yes":
                    save_changes(interface, modified_settings, menu_state)
                    logging.info("Changes Saved")

                modified_settings.clear()
                menu = rebuild_menu_at_current_path(interface, menu_state)
                pass

            menu_state.need_redraw = True

            menu_win.erase()
            if help_win:
                help_win.erase()

            # max_help_lines = 4
            # draw_help_window(menu_win.getbegyx()[0], menu_win.getbegyx()[1], menu_win.getmaxyx()[0], max_help_lines, menu_state.current_menu, selected_index, transform_menu_path(menu_state.menu_path))

            menu_win.refresh()
            if help_win:
                help_win.refresh()

            # if len(menu_state.menu_path) < 2:
            #     modified_settings.clear()

            # Navigate back to the previous menu
            if len(menu_state.menu_path) > 1:
                menu_state.menu_path.pop()
                menu_state.current_menu = menu["Main Menu"]
                for step in menu_state.menu_path[1:]:
                    menu_state.current_menu = menu_state.current_menu.get(step, {})
                menu_state.selected_index = menu_state.menu_index.pop()
                menu_state.start_index.pop()

        elif key == 27:  # Escape key
            menu_win.erase()
            menu_win.refresh()
            break


def rebuild_menu_at_current_path(interface, menu_state):
    """Rebuild menus from the device and re-point current_menu to the same path."""
    new_menu = generate_menu_from_protobuf(interface)
    cur = new_menu["Main Menu"]
    for step in menu_state.menu_path[1:]:
        cur = cur.get(step, {})
    menu_state.current_menu = cur
    return new_menu


def set_region(interface: object) -> None:
    node = interface.getNode("^local")
    device_config = node.localConfig
    lora_descriptor = device_config.lora.DESCRIPTOR

    # Get the enum mapping of region names to their numerical values
    region_enum = lora_descriptor.fields_by_name["region"].enum_type
    region_name_to_number = {v.name: v.number for v in region_enum.values}

    regions = list(region_name_to_number.keys())

    new_region_name = get_list_input(
        t("ui.prompt.select_region", default="Select your region:"), "UNSET", regions
    )

    # Convert region name to corresponding enum number
    new_region_number = region_name_to_number.get(new_region_name, 0)  # Default to 0 if not found

    node.localConfig.lora.region = new_region_number
    node.writeConfig("lora")
